//----------------------------------------------------------------------------------------------
// BSP : Seeed nRF52 Borads 1.1.1
// Board : Seeed nRF52 Borads / Seeed XIAO nRF52840 Sense 
//----------------------------------------------------------------------------------------------

// 2023/06/10
// Important : NVIC_SetPriority(RTC2_IRQn,3);

#include <Wire.h>
#include "SparkFunBME280.h"
#include <Arduino.h>
#include <Adafruit_TinyUSB.h>   // For Serial.print()
#include <bluefruit.h>          // /Arduino15/packages/Seeeduino/hardware/nrf52/1.1.1/libraries/Bluefruit52Lib/src
#include <Adafruit_SPIFlash.h>  // Need to be deleted /Documents/Arduino/libraries/SdFat

#define HICHG         22        // Charge current　Hi:50mA Lo:100mA       D22 (13u)
#define CHG           23        // Charge LED Lo:ON                       D23 (17u)
#define VBAT_ENABLE   14        // Battery voltage read enable  Lo:enable D14 (14u)   
#define VBAT_READ     32        // Battery voltage read pin               D32 (31u)
#define FILE_NAME "log.csv"     // For the datalogging file:
#define SLEEP_TIME  3600        // Sleep time [sec]
//#define SLEEP_TIME  10        // Sleep time [sec] for debug

// Built from the P25Q16H datasheet.
// https://gitlab.com/arduino5184213/seeed_xiao_nrf52840/flash_speedtest/-/tree/master
SPIFlash_Device_t const P25Q16H {
  .total_size = (1UL << 21), // 2MiB
  .start_up_time_us = 10000, // Don't know where to find that value
  
  .manufacturer_id = 0x85,
  .memory_type = 0x60,
  .capacity = 0x15,

  .max_clock_speed_mhz = 55,
  .quad_enable_bit_mask = 0x02, // Datasheet p. 27
  .has_sector_protection = 1,   // Datasheet p. 27
  .supports_fast_read = 1,      // Datasheet p. 29
  .supports_qspi = 1,           // Obviously
  .supports_qspi_writes = 1,    // Datasheet p. 41
  .write_status_register_split = 1, // Datasheet p. 28
  .single_status_byte = 0,      // 2 bytes
  .is_fram = 0,                 // Flash Memory
};

Adafruit_FlashTransport_QSPI flashTransport;
Adafruit_SPIFlash flash(&flashTransport);
FatVolume fatfs;  // file system object from SdFat

BME280 mySensor;

uint32_t mS;                  // Measurement time
float Vbatt;                  // Battery voltage [V]
float Humidity;
float Pressure;
float Temperature;
bool intFlag = true;          // Interrupu flag

//*********************************************************************************************
void setup()
{
  // Serial port initialization
  Serial.begin(115200);
//  while (!Serial);
  delay(2000);
  
  // Pin initialization
  pinMode(VBAT_READ, INPUT); 
  pinMode(HICHG, OUTPUT);
  pinMode(VBAT_ENABLE, OUTPUT);   
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_BLUE, OUTPUT);

  digitalWrite(HICHG, LOW);         // Charge current 100mA
  digitalWrite(VBAT_ENABLE, LOW);   // Battery voltage read enable   
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);

  // ADC initialization　See wireing_analog_nRF52.c:73
  analogReference(AR_DEFAULT);      // Default 0.6V*6=3.6V
  analogReadResolution(12);         // wireing_analog_nRF52.c:39
  analogSampleTime(10);             // default 3uS　3,5,10,15,20,40
    
  // BME280 Initialization
  Wire.begin();
  Wire.setClock(400000);            //Increase to fast I2C speed!

  mySensor.setI2CAddress(0x76);
  if (mySensor.beginI2C() == false) //Begin communication over I2C
  {
    Serial.println("The sensor did not respond. Please check wiring.");
    while(1); //Freeze
  }
  mySensor.setFilter(0); // 0 to 4 is valid. Filter coefficient. See 3.4.4
  mySensor.setStandbyTime(0); //0 to 7 valid. Time between readings. See table 27.
  mySensor.setTempOverSample(1); //0 to 16 are valid. 0 disables temp sensing. See table 24.
  mySensor.setPressureOverSample(1); //0 to 16 are valid. 0 disables pressure sensing. See table 23.
  mySensor.setHumidityOverSample(1); //0 to 16 are valid. 0 disables humidity sensing. See table 19.   
  Serial.println("BME280 Initialization end");

  // Enable DC-DC converter Mode
  NRF_POWER->DCDCEN = 1;            // Enable DC/DC converter for REG1 stage

  // enter Deep Power-Down Mode, See "P25Q16H Datasheet 10.1 Commands listing"
  Serial.println("enter Deep Power-down");
  flashTransport.begin();
  while(!flashTransport.runCommand(0xB9));
  delayMicroseconds(5);             // tDP=3uS
  flashTransport.end();  

  // RTC initialization  
  initRTC(1024 * SLEEP_TIME);       // SLEEP_MTIME [sec]  

  Serial.println("End of initialization");
}

//************************************************************************************************************
void loop()
{
  if(intFlag == true)     // If wake up 76mS
  {       
    digitalWrite(LED_GREEN, LOW);
 
    // BME280 start measurement
    mySensor.setMode(MODE_FORCED);
    while(mySensor.isMeasuring() == false) ; //Wait for sensor to start measurment

    // Measurement time
    mS = millis();

    // Battery voltage averaging
    int Vadc = 0;
    for (int i = 0; i < 16; i++) {
      Vadc = Vadc + analogRead(VBAT_READ);
    }
    Vadc = Vadc / 16;
    Vbatt = 2.961 * 3.6 * Vadc / 4096;   // Vref=3.6  ratio=2.961

    // release Deep Power-Down Mode
    Serial.println("release Deep Power-down");
    flashTransport.begin();
    while(!flashTransport.runCommand(0xAB));  // 0xAB release Deep Power-down                                    
    delayMicroseconds(10);            // tRES=8uS

    // Initialize flash library and check its chip ID.
    if (!flash.begin(&P25Q16H, 1)) {
      Serial.println("Error, failed to initialize flash chip!");
      while(1);
    }
    Serial.print("Flash chip JEDEC ID: 0x"); Serial.println(flash.getJEDECID(), HEX);

    // First call begin to mount the filesystem.  Check that it returns true
    // to make sure the filesystem was mounted.
    if (!fatfs.begin(&flash)) {
      Serial.println("Error, failed to mount newly formatted filesystem!");
      Serial.println("Was the flash chip formatted with the fatfs_format example?");
      while(1);
    }
    Serial.println("Mounted filesystem!");

    // log file open
    File32 dataFile = fatfs.open(FILE_NAME, FILE_WRITE);
    Serial.println("file open");
    if (dataFile)
    {
      
      // Sensor data
      Serial.println("BME280 data read");
      while(mySensor.isMeasuring() == true) ; // While sensor completes the reading  7mS
      Humidity = mySensor.readFloatHumidity();
      Pressure = mySensor.readFloatPressure();
      Temperature = mySensor.readTempC();

      dataFile.print(mS);
      dataFile.print(",");
      dataFile.print(Humidity, 3);
      dataFile.print(",");
      dataFile.print(Pressure, 3);
      dataFile.print(",");
      dataFile.print(Temperature, 3);
      dataFile.print(",");      
      dataFile.println(Vbatt, 3);

      Serial.print(mS);
      Serial.print(",  ");
      Serial.print(Humidity, 3);
      Serial.print(",  ");
      Serial.print(Pressure, 3);
      Serial.print(",  ");
      Serial.print(Temperature, 3);
      Serial.print(",  ");      
      Serial.println(Vbatt, 3);

      // file close, fatfs end, flash end
      Serial.println("file close, fatfs end, flash end");
      dataFile.close();   // 65mS
      fatfs.end();
      flash.end();

      digitalWrite(LED_GREEN, HIGH);
    }
    else
    {
      Serial.println("Failed to open data file for writing!");
    }
    // enter Deep Power-Down Mode
    Serial.println("enter Deep Power-down");
    flashTransport.runCommand(0xB9);
    delayMicroseconds(5);             // tDP=3uS
    flashTransport.end();

    Serial.println();

    intFlag = false;      
  }   
  
  // Sleep
  __WFI();
  __SEV();
  __WFI();

}

//**************************************************************************************************************
// RTC initialization
void initRTC(unsigned long RTCcount) 
{
  // See datasheet "6.22 RTC - Real-time counter 6.22.10 Registers"
  NRF_CLOCK->TASKS_LFCLKSTOP = 1;
  NRF_CLOCK->LFCLKSRC = 1;       // select LFXO
  NRF_CLOCK->TASKS_LFCLKSTART = 1;
  while(NRF_CLOCK->LFCLKSTAT != 0x10001)
  
  NRF_RTC2->TASKS_STOP = 1;      // stop counter 
  NRF_RTC2->TASKS_CLEAR = 1;     // clear counter
  NRF_RTC2->PRESCALER = 31;      // set counter prescaler, fout=32768/(1+PRESCALER)　1024Hz
  NRF_RTC2->CC[0] = RTCcount;    // set value for TRC compare register 0
  NRF_RTC2->INTENSET = 0x10000;  // enable interrupt CC[0] compare match event
  NRF_RTC2->EVTENCLR = 0x10000;  // clear counter when CC[0] event
  NVIC_SetPriority(RTC2_IRQn,3); // Higher priority than SoftDeviceAPI
  NVIC_EnableIRQ(RTC2_IRQn);     // enable interrupt  
  NRF_RTC2->TASKS_START = 1;     // start Timer
}

//**************************************************************************************************************
// RTC interrupt handler
extern "C" void RTC2_IRQHandler(void)
{
  if ((NRF_RTC2->EVENTS_COMPARE[0] != 0) && ((NRF_RTC2->INTENSET & 0x10000) != 0)) {
    NRF_RTC2->EVENTS_COMPARE[0] = 0;  // clear compare register 0 event
    NRF_RTC2->TASKS_CLEAR = 1;        // clear counter

    intFlag = true;     
  }
}
